/*
 * $Header: /home/cvs/jakarta-slide/src/share/org/apache/slide/lock/Lock.java,v 1.10 2001/09/08 19:50:37 cmlenz Exp $
 * $Revision: 1.10 $
 * $Date: 2001/09/08 19:50:37 $
 *
 * ====================================================================
 *
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights 
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:  
 *       "This product includes software developed by the 
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Tomcat", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written 
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * [Additional notices, if required by prior licensing conditions]
 *
 */ 

package org.apache.slide.lock;

import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;
import java.util.Date;
import java.util.Stack;
import org.apache.slide.common.*;
import org.apache.slide.structure.*;
import org.apache.slide.security.*;
import org.apache.slide.util.Configuration;
import org.apache.slide.authenticate.CredentialsToken;

/**
 * Lock helper class.
 * 
 * @author <a href="mailto:remm@apache.org">Remy Maucherat</a>
 * @version $Revision: 1.10 $
 */
public interface Lock {
    
    
    // ------------------------------------------------------ Interface Methods
    
    
    /**
     * Put a lock on a subject.
     * 
     * @param crendentialsToken Credentials token
     * @param token Object containing all the lock information
     * @exception ServiceAccessException Low level service access exception
     * @exception ObjectNotFoundException One of the objects referenced 
     * in the lock token were not found
     * @exception ObjectIsAlreadyLockedException Object is already locked 
     * with an incompatible lock token
     * @exception AccessDeniedException Insufficient credentials to allow 
     * object locking
     */
    void lock(SlideToken slideToken, NodeLock token)
        throws ServiceAccessException, ObjectIsAlreadyLockedException, 
        AccessDeniedException, ObjectNotFoundException;
    
    
    /**
     * Removes a lock.
     * 
     * @param crendentialsToken Credentials token
     * @param token Object containing all the lock information
     * @exception ServiceAccessException Low level service access exception
     * @exception LockTokenNotFoundException Cannot find the Lock in the 
     * Lock Store service
     */
    void unlock(SlideToken slideToken, NodeLock token)
        throws ServiceAccessException, LockTokenNotFoundException;
    
    
    /**
     * Removes a set of linked locks.
     * 
     * @param crendentialsToken Credentials token
     * @param token Object containing all the lock information
     * @exception ServiceAccessException Low level service access exception
     * @exception LockTokenNotFoundException Cannot find the Lock in the 
     * Lock Store service
     * @exception ObjectNotFoundException One of the objects referenced 
     * in the lock token were not found
     */
    void unlock(SlideToken slideToken, String objectUri, 
                String lockId)
        throws ServiceAccessException, LockTokenNotFoundException, 
        ObjectNotFoundException;
    
    
    /**
     * Renew a lock.
     * 
     * @param crendentialsToken Credentials token
     * @param token Object containing the lock info, as well as the new 
     * desired expiration date
     * @exception ServiceAccessException Low level service access exception
     * @exception LockTokenNotFoundException Cannot find the Lock in the 
     * Lock Store service
     */
    void renew(SlideToken slideToken, NodeLock token, 
               Date newExpirationDate)
        throws ServiceAccessException, LockTokenNotFoundException;
    
    
    /**
     * Renew a set of linked locks.
     * 
     * @param crendentialsToken Credentials token
     * @param token Object containing the lock info, as well as the new 
     * desired expiration date
     * @exception ServiceAccessException Low level service access exception
     * @exception LockTokenNotFoundException Cannot find the Lock in the 
     * Lock Store service
     * @exception ObjectNotFoundException One of the objects referenced 
     * in the lock token were not found
     */
    void renew(SlideToken slideToken, String objectUri, 
               String lockId, Date newExpirationDate)
        throws ServiceAccessException, LockTokenNotFoundException, 
        ObjectNotFoundException;
    
    
    /**
     * Kills locks.
     * 
     * @param crendentialsToken Credentials token
     * @param subject Subject on which locks are to be removed
     * @exception ServiceAccessException Low level service access exception
     * @exception ObjectNotFoundException One of the objects referenced 
     * in the lock token were not found
     * @exception LockTokenNotFoundException Cannot find the Lock in the 
     * Lock Store service
     * @exception AccessDeniedException Insufficient credentials to allow 
     * forced removal of locks
     */
    void kill(SlideToken slideToken, SubjectNode subject)
        throws ServiceAccessException, AccessDeniedException, 
        LockTokenNotFoundException, ObjectNotFoundException;
    
    
    /**
     * Enumerate locks on a subject, including any locks inherited from parent
     * subjects.
     * 
     * @param slideToken the slide token
     * @param objectUri the URI of the object of which we want to enumerate 
     *        active locks
     * 
     * @return an Enumeration of NodeLock objects
     * 
     * @exception ServiceAccessException low level service access exception
     * @exception ObjectNotFoundException one of the objects referenced in 
     *            the lock token were not found
     */
    Enumeration enumerateLocks(SlideToken slideToken, 
                               String objectUri)
        throws ServiceAccessException, ObjectNotFoundException, 
        LockTokenNotFoundException;
    
    
    /**
     * Enumerate locks on a subject, optionally excluding locks inherited 
     * from parent subjects.
     * 
     * @param slideToken the slide token
     * @param objectUri the URI of the object of which we want to enumerate 
     *        active locks
     * @param inherited if true, locks inherited from parent objects will be 
     *                  included in the result, otherwise, only locks that
     *                  explicitly lock the object will be returned
     * 
     * @return an Enumeration of NodeLock objects
     * 
     * @exception ServiceAccessException low level service access exception
     * @exception ObjectNotFoundException one of the objects referenced in 
     *            the lock token were not found
     * @exception LockTokenNotFoundException cannot find the lock in the 
     *            lock store service
     */
    Enumeration enumerateLocks(SlideToken slideToken, String objectUri,
                               boolean inherited)
        throws ServiceAccessException, ObjectNotFoundException, 
        LockTokenNotFoundException;
    
    
    /**
     * Utility function for lock checking. Mirrors Security.checkCredentials.
     * 
     * @param slideToken Credetials token
     * @param subject Object on which the action is performed
     * @param action Action to test
     * @exception ObjectLockedException Can't perform specified action 
     * on object
     * @exception ServiceAccessException Low level service access exception
     * @exception ObjectNotFoundException One of the objects referenced 
     * in the lock token were not found
     */
    void checkLock(SlideToken slideToken, ObjectNode object, 
                   ActionNode action)
        throws ServiceAccessException, ObjectNotFoundException, 
        ObjectLockedException;
    
    
    /**
     * Check locking for a specific action and credential.
     * 
     * @param slideToken Credetials token
     * @param subject Subject to test
     * @param user User to test
     * @param action Action to test
     * @exception ObjectLockedException Can't perform specified action 
     * on object
     * @exception ServiceAccessException Low level service access exception
     * @exception ObjectNotFoundException One of the objects referenced 
     * in the lock token were not found
     */
    void checkLock(SlideToken slideToken, ObjectNode subject, 
                   SubjectNode user, ActionNode action)
        throws ServiceAccessException, ObjectNotFoundException, 
        ObjectLockedException;
    
    
    /**
     * Tests if an element is locked.
     * 
     * @param crendentialsToken Credentials token
     * @param subject Subject to test
     * @param user User to test
     * @param action Locked for action
     * @return boolean True if the subject is locked
     * @exception ServiceAccessException Low level service access exception
     * @exception ObjectNotFoundException One of the objects referenced 
     * in the lock token were not found
     */
    boolean isLocked(SlideToken slideToken, ObjectNode subject, 
                     SubjectNode user, ActionNode action, boolean tryToLock)
        throws ServiceAccessException, ObjectNotFoundException;
    
    
    /**
     * Tests if an element is locked.
     * 
     * @param crendentialsToken Credentials token
     * @param subject Subject to test
     * @param user User to test
     * @param action Locked for action
     * @param inheritance Set to true if we want to check if any children 
     * is locked
     * with an incompatible lock
     * @param tryToLock True if the check is intended to check whether 
     * or not we can put a new lock
     * @return boolean True if the subject is locked
     * @exception ServiceAccessException Low level service access exception
     * @exception ObjectNotFoundException One of the objects referenced in 
     * the lock token were not found
     */
    boolean isLocked(SlideToken slideToken, ObjectNode subject, 
                     SubjectNode user, ActionNode action, boolean inheritance,
                     boolean tryToLock)
        throws ServiceAccessException, ObjectNotFoundException;
    
    
    /**
     * Tests if an element is locked.
     * 
     * @param token Lock token to test
     * @return boolean True if locked
     * @exception ServiceAccessException Low level service access exception
     * @exception ObjectNotFoundException One of the objects referenced 
     * in the lock token were not found
     */
    boolean isLocked(SlideToken slideToken, NodeLock token, 
                     boolean tryToLock)
        throws ServiceAccessException, ObjectNotFoundException;
    
}
